@@ -1,5 +1,5 @@
 // -*- c++ -*-
-/* $Id: dispatcher.cc,v 1.4 2003/07/27 12:02:54 murrayc Exp $ */
+/* $Id: dispatcher.cc,v 1.6 2004/06/03 21:59:53 daniel Exp $ */
 
 /* Copyright 2002 The gtkmm Development Team
  *
@@ -23,12 +23,10 @@
 #include <glibmm/fileutils.h>
 #include <glibmm/main.h>
 #include <glibmm/thread.h>
-#include <sigc++/class_slot.h>
 
-#include <fcntl.h>
 #include <cerrno>
+#include <fcntl.h>
 #include <glib.h>
-#include <algorithm> //Needed by MSVC++.Net
 
 #ifndef G_OS_WIN32
 #include <unistd.h>
@@ -37,92 +35,38 @@
 #include <io.h>
 #include <direct.h>
 #include <list>
-#include <iostream>
 #endif /* G_OS_WIN32 */
 
 
 namespace
 {
 
-#ifdef G_OS_WIN32
-/*
- * Template Class that is used to transfer notification data between
- * running thread and GUI-thread
- */
- //TODO: What does "TQ" mean?
-template<class TQ> class Locked_Queue
-{
-public:
-  Locked_Queue()
-  {
-  }
-  
-  ~Locked_Queue()
-  {
-    // We used new() to create the notification data so here we are
-    // deleting all remaining pointers in the queue.
-    std::transform(lqueue_.begin(), lqueue_.end(), lqueue_.begin(), delete_ptr);
-  }
-
-  TQ* get()
-  {
-    Glib::Mutex::Lock lock(mutex_);
-
-    TQ* ret = 0;
-    if(lqueue_.size() > 0)
-    {
-      ret = lqueue_.front();
-      lqueue_.pop_front();
-    }
-    
-    return ret;
-  }
-  
-  void put(TQ* e)
-  {
-    Glib::Mutex::Lock lock(mutex_);
-    lqueue_.push_back(e);
-  }
-  
-private:
-  Glib::Mutex mutex_;
-  std::list<TQ*> lqueue_;
-  
-  static TQ* delete_ptr(TQ* p)
-  {
-    delete p;
-    return 0;
-  }
-};
-
-/*
- * Here we define differences between Win32 and UNIX pipe types and invalid values:
- */
-#define FD_TYPE HANDLE
-#define INVALID_FD (0)
-
-#else //#ifdef G_OS_WIN32
-
-#define FD_TYPE int
-#define INVALID_FD (-1)
-
-#endif //#ifdef G_OS_WIN32
-
 struct DispatchNotifyData
 {
   unsigned long           tag;
   Glib::Dispatcher*       dispatcher;
   Glib::DispatchNotifier* notifier;
+
+  DispatchNotifyData()
+    : tag (0), dispatcher (0), notifier (0) {}
+
+  DispatchNotifyData(unsigned long tag_, Glib::Dispatcher* dispatcher_, Glib::DispatchNotifier* notifier_)
+    : tag (tag_), dispatcher (dispatcher_), notifier (notifier_) {}
 };
 
 void warn_failed_pipe_io(const char* what, int err_no)
 {
-  g_critical("Error in inter-thread communication: %s() failed: %s", what, g_strerror(err_no));
+#ifdef G_OS_WIN32
+  const char *const message = g_win32_error_message(err_no);
+#else
+  const char *const message = g_strerror(err_no);
+#endif
+  g_critical("Error in inter-thread communication: %s() failed: %s", what, message);
 }
 
 #ifndef G_OS_WIN32
-
-/* Try to set the close-on-exec flag of the file descriptor,
+/*
+ * Try to set the close-on-exec flag of the file descriptor,
  * so that it won't be leaked if a new process is spawned.
  */
 void fd_set_close_on_exec(int fd)
@@ -132,32 +76,40 @@ void fd_set_close_on_exec(int fd)
 
   fcntl(fd, F_SETFD, flags | FD_CLOEXEC);
 }
+#endif /* !G_OS_WIN32 */
 
-#endif /* G_OS_WIN32 */
-
-/* One word: paranoia.
+/*
+ * One word: paranoia.
  */
-void fd_close_and_invalidate(FD_TYPE& fd)
-{
-  if(fd != INVALID_FD)
-  {
 #ifdef G_OS_WIN32
-    // The way Win32 closes "pipe" handles
-    if( !CloseHandle(fd) )
-      warn_failed_pipe_io( "close", GetLastError() );
-#else
+void fd_close_and_invalidate(HANDLE& fd)
+{
+  if(fd != 0)
+  {
+    if(!CloseHandle(fd))
+      warn_failed_pipe_io("CloseHandle", GetLastError());
+
+    fd = 0;
+  }
+}
+#else /* !G_OS_WIN32 */
+void fd_close_and_invalidate(int& fd)
+{
+  if(fd >= 0)
+  {
     int result;
 
-    do { result = close(fd); }
+    do
+      result = close(fd);
     while(result < 0 && errno == EINTR);
 
     if(result < 0)
       warn_failed_pipe_io("close", errno);
-#endif /* G_OS_WIN32 */
 
-    fd = INVALID_FD;
+    fd = -1;
   }
 }
+#endif /* !G_OS_WIN32 */
 
 } // anonymous namespace
 
@@ -185,14 +137,15 @@ private:
 
   Glib::RefPtr<MainContext> context_;
   int                       ref_count_;
-  FD_TYPE                   fd_receiver_;
 #ifdef G_OS_WIN32
-  // queue used instead of sender pipe channel
-  Locked_Queue<DispatchNotifyData>  notify_queue_;
+  HANDLE                        fd_receiver_;
+  Glib::Mutex                   mutex_;
+  std::list<DispatchNotifyData> notify_queue_;
 #else
-  FD_TYPE                   fd_sender_;
-#endif /* G_OS_WIN32 */
-  SigC::Connection          conn_io_handler_;
+  int                       fd_receiver_;
+  int                       fd_sender_;
+#endif /* !G_OS_WIN32 */
+  sigc::connection          conn_io_handler_;
 
   void create_pipe();
   bool pipe_io_handler(Glib::IOCondition condition);
@@ -212,10 +165,14 @@ DispatchNotifier::DispatchNotifier(const Glib::RefPtr<MainContext>& context)
 :
   context_      (context),
   ref_count_    (0),
-  fd_receiver_  (INVALID_FD)
-#ifndef G_OS_WIN32
-  ,fd_sender_    (INVALID_FD)
-#endif /* G_OS_WIN32 */
+#ifdef G_OS_WIN32
+  fd_receiver_  (0),
+  mutex_        (),
+  notify_queue_ ()
+#else
+  fd_receiver_  (-1),
+  fd_sender_    (-1)
+#endif
 {
   create_pipe();
 
@@ -223,19 +180,19 @@ DispatchNotifier::DispatchNotifier(const Glib::RefPtr<MainContext>& context)
   {
 #ifdef G_OS_WIN32
     conn_io_handler_ = context_->signal_io().connect(
-        SigC::slot_class(*this, &DispatchNotifier::pipe_io_handler),
+        sigc::mem_fun(*this, &DispatchNotifier::pipe_io_handler),
         GPOINTER_TO_INT(fd_receiver_), Glib::IO_IN);
-#else
+#else /* !G_OS_WIN32 */
     conn_io_handler_ = context_->signal_io().connect(
-        SigC::slot_class(*this, &DispatchNotifier::pipe_io_handler),
+        sigc::mem_fun(*this, &DispatchNotifier::pipe_io_handler),
         fd_receiver_, Glib::IO_IN);
-#endif /* G_OS_WIN32 */
+#endif /* !G_OS_WIN32 */
   }
   catch(...)
   {
 #ifndef G_OS_WIN32
     fd_close_and_invalidate(fd_sender_);
-#endif /* G_OS_WIN32 */
+#endif /* !G_OS_WIN32 */
     fd_close_and_invalidate(fd_receiver_);
 
     throw;
@@ -244,12 +201,12 @@ DispatchNotifier::DispatchNotifier(const Glib::RefPtr<MainContext>& context)
 
 DispatchNotifier::~DispatchNotifier()
 {
-  // Disconnect manually because we're using SigC::slot_class().
+  // Disconnect manually because we don't inherit from sigc::trackable
   conn_io_handler_.disconnect();
 
 #ifndef G_OS_WIN32
   fd_close_and_invalidate(fd_sender_);
-#endif /* G_OS_WIN32 */
+#endif /* !G_OS_WIN32 */
   fd_close_and_invalidate(fd_receiver_);
 }
 
@@ -257,25 +214,25 @@ void DispatchNotifier::create_pipe()
 {
 #ifdef G_OS_WIN32
   // On Win32 we are using synchronization object instead of pipe
-  // thus storing its handle as fd_receiver_
-  if( !(fd_receiver_ = CreateEvent(NULL, FALSE, FALSE, NULL)) )
-  {
-    gint err_no = GetLastError();
-    GError* const error = g_error_new( G_FILE_ERROR, g_file_error_from_errno( err_no ),
-      "Failed to create pipe for inter-thread communication: %s",
-      g_win32_error_message( err_no ) );
+  // thus storing its handle as fd_receiver_.
+  fd_receiver_ = CreateEvent(0, FALSE, FALSE, 0);
 
+  if(!fd_receiver_)
+  {
+    const int err_no = GetLastError();
+    GError* const error = g_error_new(G_FILE_ERROR, g_file_error_from_errno(err_no),
+                                      "Failed to create pipe for inter-thread communication: %s",
+                                      g_win32_error_message(err_no));
     throw Glib::FileError(error);
   }
-#else
-  int filedes[2] = { INVALID_FD, INVALID_FD };
+#else /* !G_OS_WIN32 */
+  int filedes[2] = { -1, -1 };
 
   if(pipe(filedes) < 0)
   {
     GError* const error = g_error_new(G_FILE_ERROR, g_file_error_from_errno(errno),
-      "Failed to create pipe for inter-thread communication: %s",
-      g_strerror(errno));
-
+                                      "Failed to create pipe for inter-thread communication: %s",
+                                      g_strerror(errno));
     throw Glib::FileError(error);
   }
 
@@ -284,7 +241,7 @@ void DispatchNotifier::create_pipe()
 
   fd_receiver_ = filedes[0];
   fd_sender_   = filedes[1];
-#endif /* G_OS_WIN32 */
+#endif /* !G_OS_WIN32 */
 }
 
 // static
@@ -328,26 +285,23 @@ void DispatchNotifier::unreference_instance(DispatchNotifier* notifier)
 void DispatchNotifier::send_notification(Dispatcher* dispatcher)
 {
 #ifdef G_OS_WIN32
-  DispatchNotifyData* data = new DispatchNotifyData();
-
-  data->tag        = 0xdeadbeef;
-  data->dispatcher = dispatcher;
-  data->notifier   = this;
-
-  // put notification data into queue
-  notify_queue_.put(data);
-
-  // send notification event to GUI-thread
-  if( !SetEvent(fd_receiver_) )
   {
-    warn_failed_pipe_io("write", GetLastError());
+    Glib::Mutex::Lock lock (mutex_);
+    notify_queue_.push_back(DispatchNotifyData(0xdeadbeef, dispatcher, this));
+  }
+
+  // Send notification event to GUI-thread.
+  if(!SetEvent(fd_receiver_))
+  {
+    warn_failed_pipe_io("SetEvent", GetLastError());
     return;
   }
-#else
-  DispatchNotifyData data = { 0xdeadbeef, dispatcher, this };
+#else /* !G_OS_WIN32 */
+  DispatchNotifyData data (0xdeadbeef, dispatcher, this);
   gssize n_written;
 
-  do { n_written = write(fd_sender_, &data, sizeof(data)); }
+  do
+    n_written = write(fd_sender_, &data, sizeof(data));
   while(n_written < 0 && errno == EINTR);
 
   if(n_written < 0)
@@ -362,39 +316,43 @@ void DispatchNotifier::send_notification(Dispatcher* dispatcher)
   // it's safe to assume immediate success for the tiny amount of data we're
   // writing.
   g_return_if_fail(n_written == sizeof(data));
-#endif /* G_OS_WIN32 */
+#endif /* !G_OS_WIN32 */
 }
 
 bool DispatchNotifier::pipe_io_handler(Glib::IOCondition)
 {
 #ifdef G_OS_WIN32
-  DispatchNotifyData* data;
+  DispatchNotifyData data;
 
-  // read notification data from queue
-  while( (data = notify_queue_.get()) )
+  for(;;)
   {
-    DispatchNotifyData local_data = *data;
-    // delete pointer cuz we did create it with new() operator
-    delete data;
-    g_return_val_if_fail(local_data.tag == 0xdeadbeef, true);
-    g_return_val_if_fail(local_data.notifier == this, true);
+    {
+      Glib::Mutex::Lock lock (mutex_);
 
-    // Actually, we wouldn't need the try/catch block because
-    // the Glib::Source C callback already does it for us.
-    // However, we do it anyway because the default return value
-    // is 'false', which is not what we want.
+      if(notify_queue_.empty())
+        break;
+
+      data = notify_queue_.front();
+      notify_queue_.pop_front();
+    }
+
+    g_return_val_if_fail(data.tag == 0xdeadbeef, true);
+    g_return_val_if_fail(data.notifier == this,  true);
+
+    // Actually, we wouldn't need the try/catch block because the Glib::Source
+    // C callback already does it for us.  However, we do it anyway because the
+    // default return value is 'false', which is not what we want.
     try
     {
-      local_data.dispatcher->signal_();
+      data.dispatcher->signal_();
     }
     catch(...)
     {
       Glib::exception_handlers_invoke();
     }
   }
-#else
-
-  DispatchNotifyData data = { 0, 0, 0 };
+#else /* !G_OS_WIN32 */
+  DispatchNotifyData data;
   gsize n_read = 0;
 
   do
@@ -429,8 +387,7 @@ bool DispatchNotifier::pipe_io_handler(Glib::IOCondition)
   {
     Glib::exception_handlers_invoke();
   }
-
-#endif /* G_OS_WIN32 */
+#endif /* !G_OS_WIN32 */
 
   return true;
 }
@@ -465,7 +422,7 @@ void Dispatcher::operator()()
   emit();
 }
 
-SigC::Connection Dispatcher::connect(const SigC::Slot0<void>& slot)
+sigc::connection Dispatcher::connect(const sigc::slot<void>& slot)
 {
   return signal_.connect(slot);
 }
